@@ -1,10 +1,16 @@
+# Copyright (c) 2021 The Regents of the University of Michigan
+# All rights reserved.
+# This software is licensed under the BSD 3-Clause License.
 """Defines a polygon."""
 
+import warnings
+
 import numpy as np
 import rowan
 
-from ..bentley_ottmann import poly_point_isect
-from ..polytri import polytri
+from ..extern.bentley_ottmann import poly_point_isect
+from ..extern.polyhedron.polygon import Polygon as PolyInside
+from ..extern.polytri import polytri
 from .base_classes import Shape2D
 from .circle import Circle
 from .utils import _generate_ax, rotate_order2_tensor, translate_inertia_tensor
@@ -41,7 +47,7 @@ def _align_points_by_normal(normal, points):
     # translation we need to consider mapping both the vector and its opposite
     # (which defines an oriented coordinate system).
     rotation, _ = rowan.mapping.kabsch([normal, -normal], [[0, 0, 1], [0, 0, -1]])
-    return np.dot(points, rotation.T)
+    return np.dot(points, rotation.T), rotation
 
 
 def _is_simple(vertices):
@@ -56,7 +62,7 @@ def _is_simple(vertices):
 
 
 class Polygon(Shape2D):
-    """A simple (i.e. non-self-overlapping) polygon.
+    """A simple (non-self-overlapping) polygon.
 
     The polygon is embedded in 3-dimensions, so the normal
     vector determines which way is "up".
@@ -96,7 +102,7 @@ class Polygon(Shape2D):
         >>> triangle = coxeter.shapes.Polygon([[-1, 0], [0, 1], [1, 0]])
         >>> import numpy as np
         >>> assert np.isclose(triangle.area, 1.0)
-        >>> bounding_circle = triangle.bounding_circle
+        >>> bounding_circle = triangle.minimal_bounding_circle
         >>> assert np.isclose(bounding_circle.radius, 1.0)
         >>> assert np.allclose(triangle.center, [0., 1. / 3., 0.])
         >>> circumcircle = triangle.circumcircle
@@ -166,7 +172,7 @@ class Polygon(Shape2D):
                 raise ValueError("Not all vertices are coplanar.")
 
         if test_simple:
-            planar_vertices = _align_points_by_normal(self._normal, self._vertices)
+            planar_vertices, _ = _align_points_by_normal(self._normal, self._vertices)
             if not _is_simple(planar_vertices):
                 raise ValueError(
                     "The vertices must be passed in counterclockwise order. "
@@ -175,70 +181,10 @@ class Polygon(Shape2D):
                     "permitted."
                 )
 
-    def reorder_verts(self, clockwise=False, ref_index=0, increasing_length=True):
-        """Sort the vertices.
-
-        Sorting is done with respect to the direction of the normal vector. The
-        default ordering is counterclockwise and preserves the vertex in the
-        0th position. A different ordering may be requested; however, note that
-        clockwise ordering will result in a negative signed area of the
-        polygon.
-
-        The reordering is performed by rotating the polygon onto the :math:`xy`
-        plane, then computing the angles of all vertices. The vertices are then
-        sorted by this angle.  Note that if two points are at the same angle,
-        the ordering is arbitrary and determined by the output of
-        :func:`numpy.argsort`, which uses an unstable quicksort algorithm by
-        default.
-
-        Args:
-            clockwise (bool):
-                If True, sort in clockwise order (Default value: False).
-            ref_index (int):
-                Index indicating which vertex should be placed first in the
-                sorted list (Default value: 0).
-            increasing_length (bool):
-                If two vertices are at the same angle relative to the
-                center, when this flag is True the point closer to the center
-                comes first, otherwise the point further away comes first
-                (Default value: True).
-
-        Example:
-            >>> square = coxeter.shapes.ConvexPolygon(
-            ...   [[-1, -1], [1, -1], [-1, 1], [1, 1]])
-            >>> print(square.vertices)
-            [[-1. -1.  0.]
-             [ 1. -1.  0.]
-             [ 1.  1.  0.]
-             [-1.  1.  0.]]
-            >>> square.reorder_verts(clockwise=True)
-            >>> print(square.vertices)
-            [[-1. -1.  0.]
-             [-1.  1.  0.]
-             [ 1.  1.  0.]
-             [ 1. -1.  0.]]
-
-        """
-        verts = _align_points_by_normal(self._normal, self._vertices - self.center)
-
-        # Compute the angle of each vertex, shift so that the chosen
-        # reference_index has a value of zero, then move into the [0, 2pi]
-        # range. The secondary sorting level is in terms of distance from the
-        # origin.
-        angles = np.arctan2(verts[:, 1], verts[:, 0])
-        angles = np.mod(angles - angles[ref_index], 2 * np.pi)
-        distances = np.linalg.norm(verts, axis=1)
-        if not increasing_length:
-            distances *= -1
-        if clockwise:
-            angles = np.mod(2 * np.pi - angles, 2 * np.pi)
-        vert_order = np.lexsort((distances, angles))
-        self._vertices = self._vertices[vert_order, :]
-
     @property
     def gsd_shape_spec(self):
         """dict: Get a :ref:`complete GSD specification <shapes>`."""  # noqa: D401
-        return {"type": "Polygon", "vertices": self._vertices.tolist()}
+        return {"type": "Polygon", "vertices": self.vertices.tolist()}
 
     @property
     def normal(self):
@@ -250,6 +196,20 @@ class Polygon(Shape2D):
         """:math:`(N_{verts}, 3)` :class:`numpy.ndarray` of float: Get the vertices of the polygon."""  # noqa: E501
         return self._vertices
 
+    @property
+    def num_vertices(self):
+        """int: Get the number of vertices."""
+        return self.vertices.shape[0]
+
+    def _rescale(self, scale):
+        """Multiply length scale.
+
+        Args:
+            scale (float):
+                Scale factor.
+        """
+        self._vertices *= scale
+
     @property
     def perimeter(self):
         """float: Get the perimeter of the polygon."""
@@ -259,16 +219,25 @@ class Polygon(Shape2D):
             )
         )
 
+    @perimeter.setter
+    def perimeter(self, value):
+        if value > 0:
+            scale = value / self.perimeter
+            self._rescale(scale)
+        else:
+            raise ValueError("Perimeter must be greater than zero.")
+
     @property
     def signed_area(self):
         """float: Get the polygon's area.
 
         To support polygons embedded in 3 dimensional space, we employ a
-        projection- and rescaling-based algorithm described
-        `here <https://geomalgorithms.com/a01-_area.html>`_. Specifically, the
-        polygon is projected onto the plane it is "most parallel" to, the area
-        of the projected polygon is computed, then the area is rescaled by the
-        component of the normal in the projected dimension.
+        projection- and rescaling-based algorithm described `here
+        <https://geomalgorithms.com/a01-_area.html#3D-Planar-Polygons>`__.
+        Specifically, the polygon is projected onto the plane it is "most
+        parallel" to, the area of the projected polygon is computed, then the
+        area is rescaled by the component of the normal in the projected
+        dimension.
         """
         # Choose the dimension to project out based on the largest magnitude
         # component of the normal vector.
@@ -299,13 +268,13 @@ class Polygon(Shape2D):
         return np.abs(self.signed_area)
 
     @area.setter
-    def area(self, new_area):
-        scale_factor = np.sqrt(new_area / self.area)
-        self._vertices *= scale_factor
+    def area(self, value):
+        scale = np.sqrt(value / self.area)
+        self._rescale(scale)
 
     @property
     def planar_moments_inertia(self):
-        r"""Get the planar moments of inertia.
+        r"""list[float, float, float]: Get the planar and product moments of inertia.
 
         Moments are computed with respect to the :math:`x` and :math:`y`
         axes. In addition to the two planar moments, this property also
@@ -345,7 +314,7 @@ class Polygon(Shape2D):
         :math:`y` position) should not be relied upon.
         """  # noqa: E501
         # Rotate shape so that normal vector coincides with z-axis.
-        verts = _align_points_by_normal(self._normal, self._vertices)
+        verts, _ = _align_points_by_normal(self._normal, self._vertices)
 
         shifted_verts = np.roll(verts, shift=-1, axis=0)
 
@@ -390,34 +359,58 @@ class Polygon(Shape2D):
         # is important: we must translate before rotating so that the parallel
         # axis theorem can be applied in the reverse direction (rotating about
         # the origin before translating to the actual centroid).
-        center = self.center
+        original_center = self.center.copy()
+        original_vertices = self._vertices.copy()
+        original_normal = self._normal.copy()
+
         self.center = (0, 0, 0)
         mat, _ = rowan.mapping.kabsch(
             [self.normal, -self.normal], [[0, 0, 1], [0, 0, -1]]
         )
         self._vertices = self._vertices.dot(mat.T)
-        original_normal = self._normal
         self._normal = np.asarray([0, 0, 1])
 
         inertia_tensor = np.diag([0, 0, self.polar_moment_inertia])
         shifted_inertia_tensor = translate_inertia_tensor(
-            center, rotate_order2_tensor(mat, inertia_tensor), self.area
+            original_center, rotate_order2_tensor(mat, inertia_tensor), self.area
         )
 
-        self.center = center
-        self._vertices = self._vertices.dot(mat)
+        self.center = original_center
+        self._vertices = original_vertices
         self._normal = original_normal
 
         return shifted_inertia_tensor
 
     @property
-    def center(self):
-        """:math:`(3, )` :class:`numpy.ndarray` of float: Get or set the centroid of the shape."""  # noqa: E501
-        return np.mean(self.vertices, axis=0)
+    def centroid(self):
+        """:math:`(3, )` :class:`numpy.ndarray` of float: Get or set the centroid of the shape.
 
-    @center.setter
-    def center(self, value):
-        self._vertices += np.asarray(value) - self.center
+        The centroid of a polygon is calculated according to
+        `this formula <https://en.wikipedia.org/wiki/Centroid#Of_a_polygon>`__.
+        """  # noqa: E501
+        verts, rotation = _align_points_by_normal(self.normal, self.vertices)
+        verts_shifted = np.roll(verts, shift=-1, axis=0)
+
+        delta_term = (
+            verts[:, 0] * verts_shifted[:, 1] - verts_shifted[:, 0] * verts[:, 1]
+        )
+        c_x = np.sum((verts[:, 0] + verts_shifted[:, 0]) * delta_term)
+        c_y = np.sum((verts[:, 1] + verts_shifted[:, 1]) * delta_term)
+
+        in_plane_centroid = np.array([c_x, c_y, 0]) / (6 * self.area)
+
+        # We've rotated into the plane, so the z position of all vertices
+        # should be equal. We take the average to improve numerical stablity.
+        in_plane_centroid[2] = np.mean(verts[:, 2])
+
+        # Revert the rotation into the plane.
+        centroid = rotation.T.dot(in_plane_centroid)
+
+        return centroid
+
+    @centroid.setter
+    def centroid(self, value):
+        self._vertices += np.asarray(value) - self.centroid
 
     def _triangulation(self):
         """Generate a triangulation of the polygon.
@@ -452,7 +445,7 @@ class Polygon(Shape2D):
         """
         ax = _generate_ax(ax)
         verts = self._vertices - self.center if center else self._vertices
-        verts = _align_points_by_normal(self._normal, verts)
+        verts, _ = _align_points_by_normal(self._normal, verts)
         verts = np.concatenate((verts, verts[[0]]))
         x = verts[:, 0]
         y = verts[:, 1]
@@ -468,6 +461,17 @@ class Polygon(Shape2D):
 
     @property
     def bounding_circle(self):
+        """:class:`~.Circle`: Get the minimal bounding circle."""
+        warnings.warn(
+            "The bounding_circle property is deprecated, use "
+            "minimal_bounding_circle instead",
+            DeprecationWarning,
+        )
+
+        return self.minimal_bounding_circle
+
+    @property
+    def minimal_bounding_circle(self):
         """:class:`~.Circle`: Get the minimal bounding circle."""
         if not MINIBALL:
             raise ImportError(
@@ -508,15 +512,23 @@ class Polygon(Shape2D):
         A `circumcircle
         <https://en.wikipedia.org/wiki/Circumscribed_circle>`__ must touch
         all the points of the polygon. A circumcircle exists if and only if
-        there is a point equidistant from all the vertices.
+        there is a point equidistant from all the vertices. The circumcircle is
+        found by finding the least squares solution of the overdetermined system
+        of linear equations defined by this constraint, and the circumcircle
+        only exists if the resulting solution has no residual.
 
         Raises:
             RuntimeError: If no circumcircle exists for this polygon.
-
         """
-        # Solves a linear system of equations to find a point equidistant from
-        # all the vertices if it exists. Since the polygon is embedded in 3D,
-        # we must constrain our solutions to the plane of the polygon.
+        # The circumsphere is defined by center C and radius r. For vertex i
+        # with position r_i, dot(r_i - C, r_i - C) = r^2, which is equivalent
+        # to dot(r_i, r_i) - 2 dot(C, r_i) + dot(C, C) = r^2, a system of
+        # quadratic equations. If we choose r_0 as the origin, then dot(C, C) =
+        # r^2 and we instead have the linear equations dot(C, p_i) =
+        # dot(p_i, p_i) / 2 where p_i = r_i - r_0. This is the set of equations
+        # that we solve. The polygon is embedded in 3D, which imposes the
+        # additional constraint that the circumcircle must lie in the plane of
+        # the polygon.
         points = np.concatenate(
             (self.vertices[1:] - self.vertices[0], self.normal[np.newaxis])
         )
@@ -529,6 +541,78 @@ class Polygon(Shape2D):
 
         return Circle(np.linalg.norm(x), x + self.vertices[0])
 
+    @property
+    def circumcircle_radius(self):
+        """float: Get the radius of the polygon's circumcircle."""
+        return self.circumcircle.radius
+
+    @circumcircle_radius.setter
+    def circumcircle_radius(self, value):
+        self._rescale(value / self.circumcircle_radius)
+
+    @property
+    def incircle(self):
+        """:class:`~.Sphere`: Get the polygon's incircle.
+
+        Note:
+            The incircle of a polygon is defined as the circle contained within
+            the polygon that is tangent to all its faces. This condition
+            uniquely defines the circle, if it exists. The set of equations
+            defined by this equation is solved using a least squares approach,
+            with the magnitude of the residual used to determine whether or not
+            the incircle exists.
+
+        """
+        # The incircle is defined by center C and radius r. For face i
+        # defined by its unit normal n_i and any point in the plane (choose a
+        # vertex v_i for convenience), we must have dot(C + r n_i - v_i, n_i) = 0.
+        # Defining the vector Cr = (C_x, C_y, C_z, r), and the augmented
+        # normals m = (n_x, n_y, n_z, 1), rearranging gives the equations that
+        # we solve: dot(m_i, cr) = dot(n_i, v_i). The polygon is embedded in
+        # 3D, which imposes the additional constraint that the incircle must
+        # lie in the plane of the polygon.
+
+        outward_normals = np.cross(
+            # Order is important here to get the outward facing normal.
+            np.roll(self.vertices, axis=0, shift=-1) - self.vertices,
+            self.normal,
+        )
+        outward_normals /= np.linalg.norm(outward_normals, axis=-1)[:, np.newaxis]
+
+        # vstack the row corresponding to the constraint equation
+        a = np.vstack(
+            (
+                # hstack on the coefficient of 1 for the radius
+                np.hstack((outward_normals, np.ones((self.num_vertices, 1)))),
+                # The coefficient of the radius in the plane constraint is 0.
+                np.append(self.normal, 0),
+            )
+        )
+
+        b = np.concatenate(
+            (
+                # Arbitrarily choose the first vertex in every edge for the dot product.
+                np.sum(outward_normals * self.vertices, axis=-1),
+                # Can use the dot product of the normal with any of the vertices.
+                [np.dot(self.normal, self.vertices[0])],
+            )
+        )
+
+        x, resids, _, _ = np.linalg.lstsq(a, b, None)
+        if len(self.vertices) > 4 and not np.isclose(resids, 0):
+            raise RuntimeError("No incircle for this polygon.")
+
+        return Circle(x[3], x[:3])
+
+    @property
+    def incircle_radius(self):
+        """float: Get the radius of the polygon's incircle."""
+        return self.incircle.radius
+
+    @incircle_radius.setter
+    def incircle_radius(self, value):
+        self._rescale(value / self.incircle_radius)
+
     def compute_form_factor_amplitude(self, q, density=1.0):  # noqa: D102
         """Calculate the form factor intensity.
 
@@ -579,3 +663,50 @@ class Polygon(Shape2D):
         )
         form_factor *= density
         return form_factor
+
+    def is_inside(self, points):
+        """Determine whether points are contained in this polygon.
+
+        .. note::
+
+            Points on the boundary of the shape will return :code:`True`.
+
+        Args:
+            points (:math:`(N, 3)` :class:`numpy.ndarray`):
+                The points to test.
+
+        Returns:
+            :math:`(N, )` :class:`numpy.ndarray`:
+                Boolean array indicating which points are contained in the
+                polyhedron.
+        """
+        # Rotate both the vertices and the points into the plane.
+        verts, rotation = _align_points_by_normal(self.normal, self.vertices)
+        points_in_plane = np.dot(points, rotation.T)
+
+        winding_number_calculator = PolyInside(verts)
+
+        def _check_inside(p):
+            """Check if point is inside, including boundary points.
+
+            The polyhedron check will will raise a ValueError for points on the
+            boundary, which we want to be inside.
+            """
+            try:
+                return winding_number_calculator.winding_number(p) != 0
+            except ValueError as e:
+                if str(e) in (
+                    "vertex coincides with origin",
+                    "vertices collinear with origin",
+                ):
+                    return True
+                else:
+                    raise
+
+        return np.array([_check_inside(p) for p in np.atleast_2d(points_in_plane)])
+
+    def __repr__(self):
+        return (
+            f"coxeter.shapes.Polygon(vertices={self.vertices.tolist()}, "
+            f"normal={self.normal.tolist()})"
+        )
